# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_config import cfg
from oslo_policy import policy

ADMIN = "role:admin"
ADMIN_OR_SERVICE = "role:admin or role:service"
CREATOR = 'user:%(creator)s or project_id:%(created_by_project_id)s'
ADMIN_OR_CREATOR = 'role:admin or %s' % CREATOR
RESOURCE_OWNER = "project_id:%(project_id)s"
METRIC_OWNER = "project_id:%(resource.project_id)s"
UNPROTECTED = ""

RULE_ADMIN_OR_SERVICE = "rule:admin_or_service"
RULE_ADMIN_OR_CREATOR = "rule:admin_or_creator"
RULE_ADMIN_OR_SERVICE_OR_CREATOR = "rule:admin_or_service or rule:creator"
RULE_ADMIN_OR_CREATOR_OR_RESOURCE_OWNER = \
    "rule:admin_or_creator or rule:resource_owner"
RULE_ADMIN_OR_SERVICE_OR_CREATOR_OR_RESOURCE_OWNER = \
    "rule:admin_or_service or rule:creator or rule:resource_owner"
RULE_ADMIN_OR_CREATOR_OR_METRIC_OWNER = \
    "rule:admin_or_creator or rule:metric_owner"
RULE_ADMIN_OR_SERVICE_OR_CREATOR_OR_METRIC_OWNER = \
    "rule:admin_or_service or rule:creator or rule:metric_owner"

rules = [
    policy.RuleDefault(
        name="context_is_admin",
        check_str=ADMIN
    ),
    policy.RuleDefault(
        name="admin_or_creator",
        check_str=ADMIN_OR_CREATOR,
    ),
    policy.RuleDefault(
        name="admin_or_service",
        check_str=ADMIN_OR_SERVICE,
    ),
    policy.RuleDefault(
        name="creator",
        check_str=CREATOR,
    ),
    policy.RuleDefault(
        name="resource_owner",
        check_str=RESOURCE_OWNER
    ),
    policy.RuleDefault(
        name="metric_owner",
        check_str=METRIC_OWNER
    )
]

status_rules = [
    policy.DocumentedRuleDefault(
        name="get status",
        check_str=ADMIN,
        scope_types=['system', 'domain', 'project'],
        description='Get status of Gnocchi service.',
        operations=[
            {
                'path': '/v1/status',
                'method': 'GET'
            }
        ]
    )
]

resource_rules = [
    policy.DocumentedRuleDefault(
        name="create resource",
        check_str=UNPROTECTED,
        scope_types=['system', 'domain', 'project'],
        description='Create a new resource.',
        operations=[
            {
                'path': '/v1/resource/{resource_type}',
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="get resource",
        check_str=RULE_ADMIN_OR_CREATOR_OR_RESOURCE_OWNER,
        scope_types=['system', 'domain', 'project'],
        description='Get a resource.',
        operations=[
            {
                'path': '/v1/resource/{resource_type}/{resource_id}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="update resource",
        check_str=RULE_ADMIN_OR_SERVICE_OR_CREATOR,
        scope_types=['system', 'domain', 'project'],
        description='Update a resource.',
        operations=[
            {
                'path': '/v1/resource/{resource_type}/{resource_id}',
                'method': 'PATCH'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="delete resource",
        check_str=RULE_ADMIN_OR_CREATOR,
        scope_types=['system', 'domain', 'project'],
        description='Delete a resource.',
        operations=[
            {
                'path': '/v1/resource/{resource_type}/{resource_id}',
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="delete resources",
        check_str=RULE_ADMIN_OR_CREATOR,
        scope_types=['system', 'domain', 'project'],
        description='Delete multiple resources.',
        operations=[
            {
                'path': '/v1/resource/{resource_type}',
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="list resource",
        check_str=RULE_ADMIN_OR_CREATOR_OR_RESOURCE_OWNER,
        scope_types=['system', 'domain', 'project'],
        description='List all resources.',
        operations=[
            {
                'path': '/v1/resource/{resource_type}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="search resource",
        check_str=RULE_ADMIN_OR_SERVICE_OR_CREATOR_OR_RESOURCE_OWNER,
        scope_types=['system', 'domain', 'project'],
        description='Search resources.',
        operations=[
            {
                'path': '/v1/search/resources/{resource_type}',
                'method': 'POST'
            }
        ]
    )
]

resource_type_rules = [
    policy.DocumentedRuleDefault(
        name="create resource type",
        check_str=RULE_ADMIN_OR_SERVICE,
        scope_types=['system', 'domain', 'project'],
        description='Create a new resource type.',
        operations=[
            {
                'path': '/v1/resource_type',
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="get resource type",
        check_str=UNPROTECTED,
        scope_types=['system', 'domain', 'project'],
        description='Get a resource type.',
        operations=[
            {
                'path': '/v1/resource_type/{resource_type}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="update resource type",
        check_str=RULE_ADMIN_OR_SERVICE,
        scope_types=['system', 'domain', 'project'],
        description='Update a resource type.',
        operations=[
            {
                'path': '/v1/resource_type/{resource_type}',
                'method': 'PATCH'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="delete resource type",
        check_str=ADMIN,
        scope_types=['system', 'domain', 'project'],
        description='Delete a resource type.',
        operations=[
            {
                'path': '/v1/resource_type/{resource_type}',
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="list resource type",
        check_str=UNPROTECTED,
        scope_types=['system', 'domain', 'project'],
        description='List all resource types.',
        operations=[
            {
                'path': '/v1/resource_type',
                'method': 'GET'
            }
        ]
    )
]

archive_policy_rules = [
    policy.DocumentedRuleDefault(
        name="create archive policy",
        check_str=RULE_ADMIN_OR_SERVICE,
        scope_types=['system', 'domain', 'project'],
        description='Create a new archive policy',
        operations=[
            {
                'path': '/v1/archive_policy',
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="get archive policy",
        check_str=UNPROTECTED,
        scope_types=['system', 'domain', 'project'],
        description='Get an archive policy',
        operations=[
            {
                'path': '/v1/archive_policy/{archive_policy}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="update archive policy",
        check_str=ADMIN,
        scope_types=['system', 'domain', 'project'],
        description='Update an archive policy',
        operations=[
            {
                'path': '/v1/archive_policy/{archive_policy}',
                'method': 'PATCH'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="delete archive policy",
        check_str=ADMIN,
        scope_types=['system', 'domain', 'project'],
        description='Delete an archive policy',
        operations=[
            {
                'path': '/v1/archive_policy/{archive_policy}',
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="list archive policy",
        check_str=UNPROTECTED,
        scope_types=['system', 'domain', 'project'],
        description='List all archive policies',
        operations=[
            {
                'path': '/v1/archive_policy',
                'method': 'GET'
            }
        ]
    )
]

archive_policy_rule_rules = [
    policy.DocumentedRuleDefault(
        name="create archive policy rule",
        check_str=ADMIN,
        scope_types=['system', 'domain', 'project'],
        description='Create a new archive policy rule',
        operations=[
            {
                'path': '/v1/archive_policy_rule',
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="get archive policy rule",
        check_str=UNPROTECTED,
        scope_types=['system', 'domain', 'project'],
        description='Get an archive policy rule',
        operations=[
            {
                'path': '/v1/archive_policy_rule/{archive_policy_rule}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="update archive policy rule",
        check_str=ADMIN,
        scope_types=['system', 'domain', 'project'],
        description='Update an archive policy rule',
        operations=[
            {
                'path': '/v1/archive_policy_rule/{archive_policy_rule}',
                'method': 'PATCH'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="delete archive policy rule",
        check_str=ADMIN,
        scope_types=['system', 'domain', 'project'],
        description='Delete an archive policy rule',
        operations=[
            {
                'path': '/v1/archive_policy_rule/{archive_policy_rule}',
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="list archive policy rule",
        check_str=UNPROTECTED,
        scope_types=['system', 'domain', 'project'],
        description='List all archive policy rules',
        operations=[
            {
                'path': '/v1/archive_policy_rule',
                'method': 'GET'
            }
        ]
    )
]

metric_rules = [
    policy.DocumentedRuleDefault(
        name="create metric",
        check_str=UNPROTECTED,
        scope_types=['system', 'domain', 'project'],
        description='Create a new metric',
        operations=[
            {
                'path': '/v1/metric',
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="get metric",
        check_str=RULE_ADMIN_OR_SERVICE_OR_CREATOR_OR_METRIC_OWNER,
        scope_types=['system', 'domain', 'project'],
        description='Get a metric',
        operations=[
            {
                'path': '/v1/metric/{metric}',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="delete metric",
        check_str=RULE_ADMIN_OR_CREATOR,
        scope_types=['system', 'domain', 'project'],
        description='Delete a metric',
        operations=[
            {
                'path': '/v1/metric/{metric}',
                'method': 'DELETE'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="list metric",
        check_str=RULE_ADMIN_OR_CREATOR_OR_METRIC_OWNER,
        scope_types=['system', 'domain', 'project'],
        description='List all metrics',
        operations=[
            {
                'path': '/v1/metric',
                'method': 'GET'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="search metric",
        check_str=RULE_ADMIN_OR_CREATOR_OR_METRIC_OWNER,
        scope_types=['system', 'domain', 'project'],
        description='Search metrics',
        operations=[
            {
                'path': '/v1/search/metric',
                'method': 'POST'
            }
        ]
    )
]

measure_rules = [
    policy.DocumentedRuleDefault(
        name="post measures",
        check_str=RULE_ADMIN_OR_SERVICE_OR_CREATOR,
        scope_types=['system', 'domain', 'project'],
        description='Post measures',
        operations=[
            {
                'path': '/v1/metric/{metric}/measures',
                'method': 'POST'
            },
            {
                'path': '/v1/batch/metrics/measures',
                'method': 'POST'
            },
            {
                'path': '/v1/batch/resources/metrics/measures',
                'method': 'POST'
            }
        ]
    ),
    policy.DocumentedRuleDefault(
        name="get measures",
        check_str=RULE_ADMIN_OR_SERVICE_OR_CREATOR_OR_METRIC_OWNER,
        scope_types=['system', 'domain', 'project'],
        description='Get measures',
        operations=[
            {
                'path': '/v1/metric/{metric}/measures',
                'method': 'GET'
            }
        ]
    )
]


def list_rules():
    return rules + status_rules \
        + resource_rules + resource_type_rules \
        + archive_policy_rules + archive_policy_rule_rules \
        + metric_rules + measure_rules


def init(conf):
    policy_enforcer = policy.Enforcer(conf)
    policy_enforcer.register_defaults(list_rules())
    return policy_enforcer


def get_enforcer():
    # This method is used by oslopolicy CLI scripts in order to generate policy
    # files from overrides on disk and defaults in code.
    cfg.CONF([], project='gnocchi')
    return init(cfg.CONF)



# -*- encoding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from collections import defaultdict
import contextlib
import daiquiri
import datetime
import json
import uuid

from gnocchi.common import swift
from gnocchi import incoming
from gnocchi import utils

swclient = swift.swclient
swift_utils = swift.swift_utils

LOG = daiquiri.getLogger(__name__)


class SwiftStorage(incoming.IncomingDriver):
    # NOTE(sileht): Using threads with swiftclient doesn't work
    # as expected, so disable it
    MAP_METHOD = staticmethod(utils.sequencial_map)

    def __init__(self, conf, greedy=True):
        super(SwiftStorage, self).__init__(conf)
        self.swift = swift.get_connection(conf)

    def __str__(self):
        return self.__class__.__name__

    def _get_storage_sacks(self):
        __, data = self.swift.get_object(self.CFG_PREFIX, self.CFG_PREFIX)
        return json.loads(data)[self.CFG_SACKS]

    def set_storage_settings(self, num_sacks):
        self.swift.put_container(self.CFG_PREFIX)
        self.swift.put_object(self.CFG_PREFIX, self.CFG_PREFIX,
                              json.dumps({self.CFG_SACKS: num_sacks}))
        for sack in self.iter_sacks():
            self.swift.put_container(str(sack))

    def remove_sacks(self):
        for sack in self.iter_sacks():
            self.swift.delete_container(str(sack))

    def _store_new_measures(self, metric_id, data):
        now = datetime.datetime.utcnow().strftime("_%Y%m%d_%H:%M:%S")
        self.swift.put_object(
            str(self.sack_for_metric(metric_id)),
            str(metric_id) + "/" + str(uuid.uuid4()) + now,
            data)

    def _build_report(self, details):
        metric_details = defaultdict(int)
        nb_metrics = 0
        measures = 0
        for sack in self.iter_sacks():
            if details:
                headers, files = self.swift.get_container(
                    str(sack), full_listing=True)
                for f in files:
                    metric, __ = f['name'].split("/", 1)
                    metric_details[metric] += 1
            else:
                headers, files = self.swift.get_container(
                    str(sack), delimiter='/', full_listing=True)
                nb_metrics += len([f for f in files if 'subdir' in f])
            measures += int(headers.get('x-container-object-count'))
        return (nb_metrics or len(metric_details), measures,
                metric_details if details else None)

    def _list_measure_files_for_metric(self, sack, metric_id):
        headers, files = self.swift.get_container(
            str(sack), path=str(metric_id),
            full_listing=True)
        return files

    def delete_unprocessed_measures_for_metric(self, metric_id):
        sack = self.sack_for_metric(metric_id)
        files = self._list_measure_files_for_metric(sack, metric_id)
        swift.bulk_delete(self.swift, str(sack), files)

    def has_unprocessed(self, metric_id):
        sack = self.sack_for_metric(metric_id)
        return bool(self._list_measure_files_for_metric(sack, metric_id))

    @contextlib.contextmanager
    def process_measure_for_metrics(self, metric_ids):
        measures = {}
        all_files = defaultdict(list)
        for metric_id in metric_ids:
            sack = self.sack_for_metric(metric_id)
            sack_name = str(sack)
            files = self._list_measure_files_for_metric(sack, metric_id)
            all_files[sack_name].extend(files)
            measures[metric_id] = self._array_concatenate([
                self._unserialize_measures(
                    f['name'],
                    self.swift.get_object(sack_name, f['name'])[1],
                )
                for f in files
            ])

        yield measures

        # Now clean objects
        for sack_name, files in all_files.items():
            swift.bulk_delete(self.swift, sack_name, files)

    @contextlib.contextmanager
    def process_measures_for_sack(self, sack):
        measures = defaultdict(self._make_measures_array)
        sack_name = str(sack)
        headers, files = self.swift.get_container(sack_name, full_listing=True)
        for f in files:
            try:
                metric_id, random_id = f['name'].split("/")
                metric_id = uuid.UUID(metric_id)
            except ValueError:
                LOG.warning("Unable to parse measure file name %s", f)
                continue
            measures[metric_id] = self._array_concatenate([
                measures[metric_id],
                self._unserialize_measures(
                    metric_id,
                    self.swift.get_object(sack_name, f['name'])[1],
                )
            ])

        yield measures

        swift.bulk_delete(self.swift, sack_name, files)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: remote.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0cremote.proto\x12\x06remote\"-\n\x06Sample\x12\r\n\x05value\x18\x01 \x01(\x01\x12\x14\n\x0ctimestamp_ms\x18\x02 \x01(\x03\"(\n\tLabelPair\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t\"P\n\nTimeSeries\x12!\n\x06labels\x18\x01 \x03(\x0b\x32\x11.remote.LabelPair\x12\x1f\n\x07samples\x18\x02 \x03(\x0b\x32\x0e.remote.Sample\"6\n\x0cWriteRequest\x12&\n\ntimeseries\x18\x01 \x03(\x0b\x32\x12.remote.TimeSeries\"-\n\x0bReadRequest\x12\x1e\n\x07queries\x18\x01 \x03(\x0b\x32\r.remote.Query\"4\n\x0cReadResponse\x12$\n\x07results\x18\x01 \x03(\x0b\x32\x13.remote.QueryResult\"e\n\x05Query\x12\x1a\n\x12start_timestamp_ms\x18\x01 \x01(\x03\x12\x18\n\x10\x65nd_timestamp_ms\x18\x02 \x01(\x03\x12&\n\x08matchers\x18\x03 \x03(\x0b\x32\x14.remote.LabelMatcher\"L\n\x0cLabelMatcher\x12\x1f\n\x04type\x18\x01 \x01(\x0e\x32\x11.remote.MatchType\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\t\"5\n\x0bQueryResult\x12&\n\ntimeseries\x18\x01 \x03(\x0b\x32\x12.remote.TimeSeries*J\n\tMatchType\x12\t\n\x05\x45QUAL\x10\x00\x12\r\n\tNOT_EQUAL\x10\x01\x12\x0f\n\x0bREGEX_MATCH\x10\x02\x12\x12\n\x0eREGEX_NO_MATCH\x10\x03\x62\x06proto3')

_MATCHTYPE = DESCRIPTOR.enum_types_by_name['MatchType']
MatchType = enum_type_wrapper.EnumTypeWrapper(_MATCHTYPE)
EQUAL = 0
NOT_EQUAL = 1
REGEX_MATCH = 2
REGEX_NO_MATCH = 3


_SAMPLE = DESCRIPTOR.message_types_by_name['Sample']
_LABELPAIR = DESCRIPTOR.message_types_by_name['LabelPair']
_TIMESERIES = DESCRIPTOR.message_types_by_name['TimeSeries']
_WRITEREQUEST = DESCRIPTOR.message_types_by_name['WriteRequest']
_READREQUEST = DESCRIPTOR.message_types_by_name['ReadRequest']
_READRESPONSE = DESCRIPTOR.message_types_by_name['ReadResponse']
_QUERY = DESCRIPTOR.message_types_by_name['Query']
_LABELMATCHER = DESCRIPTOR.message_types_by_name['LabelMatcher']
_QUERYRESULT = DESCRIPTOR.message_types_by_name['QueryResult']
Sample = _reflection.GeneratedProtocolMessageType('Sample', (_message.Message,), {
  'DESCRIPTOR' : _SAMPLE,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.Sample)
  })
_sym_db.RegisterMessage(Sample)

LabelPair = _reflection.GeneratedProtocolMessageType('LabelPair', (_message.Message,), {
  'DESCRIPTOR' : _LABELPAIR,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.LabelPair)
  })
_sym_db.RegisterMessage(LabelPair)

TimeSeries = _reflection.GeneratedProtocolMessageType('TimeSeries', (_message.Message,), {
  'DESCRIPTOR' : _TIMESERIES,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.TimeSeries)
  })
_sym_db.RegisterMessage(TimeSeries)

WriteRequest = _reflection.GeneratedProtocolMessageType('WriteRequest', (_message.Message,), {
  'DESCRIPTOR' : _WRITEREQUEST,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.WriteRequest)
  })
_sym_db.RegisterMessage(WriteRequest)

ReadRequest = _reflection.GeneratedProtocolMessageType('ReadRequest', (_message.Message,), {
  'DESCRIPTOR' : _READREQUEST,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.ReadRequest)
  })
_sym_db.RegisterMessage(ReadRequest)

ReadResponse = _reflection.GeneratedProtocolMessageType('ReadResponse', (_message.Message,), {
  'DESCRIPTOR' : _READRESPONSE,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.ReadResponse)
  })
_sym_db.RegisterMessage(ReadResponse)

Query = _reflection.GeneratedProtocolMessageType('Query', (_message.Message,), {
  'DESCRIPTOR' : _QUERY,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.Query)
  })
_sym_db.RegisterMessage(Query)

LabelMatcher = _reflection.GeneratedProtocolMessageType('LabelMatcher', (_message.Message,), {
  'DESCRIPTOR' : _LABELMATCHER,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.LabelMatcher)
  })
_sym_db.RegisterMessage(LabelMatcher)

QueryResult = _reflection.GeneratedProtocolMessageType('QueryResult', (_message.Message,), {
  'DESCRIPTOR' : _QUERYRESULT,
  '__module__' : 'remote_pb2'
  # @@protoc_insertion_point(class_scope:remote.QueryResult)
  })
_sym_db.RegisterMessage(QueryResult)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _MATCHTYPE._serialized_start=588
  _MATCHTYPE._serialized_end=662
  _SAMPLE._serialized_start=24
  _SAMPLE._serialized_end=69
  _LABELPAIR._serialized_start=71
  _LABELPAIR._serialized_end=111
  _TIMESERIES._serialized_start=113
  _TIMESERIES._serialized_end=193
  _WRITEREQUEST._serialized_start=195
  _WRITEREQUEST._serialized_end=249
  _READREQUEST._serialized_start=251
  _READREQUEST._serialized_end=296
  _READRESPONSE._serialized_start=298
  _READRESPONSE._serialized_end=350
  _QUERY._serialized_start=352
  _QUERY._serialized_end=453
  _LABELMATCHER._serialized_start=455
  _LABELMATCHER._serialized_end=531
  _QUERYRESULT._serialized_start=533
  _QUERYRESULT._serialized_end=586
# @@protoc_insertion_point(module_scope)



# -*- encoding: utf-8 -*-
#
# Copyright © 2016 Red Hat, Inc.
# Copyright © 2014-2015 eNovance
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_db.sqlalchemy import models
import sqlalchemy
from sqlalchemy.ext import declarative
from sqlalchemy.orm import declarative_base
import sqlalchemy_utils

from gnocchi import archive_policy
from gnocchi import indexer
from gnocchi.indexer import sqlalchemy_types as types
from gnocchi import resource_type
from gnocchi import utils

Base = declarative_base()

COMMON_TABLES_ARGS = {'mysql_charset': "utf8",
                      'mysql_engine': "InnoDB"}


class GnocchiBase(models.ModelBase):
    __table_args__ = (
        COMMON_TABLES_ARGS,
    )


class ArchivePolicyDefinitionType(sqlalchemy_utils.JSONType):
    def process_bind_param(self, value, dialect):
        if value is not None:
            return super(
                ArchivePolicyDefinitionType, self).process_bind_param(
                    [v.serialize() for v in value],
                    dialect)

    def process_result_value(self, value, dialect):
        values = super(ArchivePolicyDefinitionType,
                       self).process_result_value(value, dialect)
        if values is None:
            return []
        return [archive_policy.ArchivePolicyItem(**v) for v in values]


class SetType(sqlalchemy_utils.JSONType):
    def process_result_value(self, value, dialect):
        values = super(SetType, self).process_result_value(value, dialect)
        if values is None:
            return set()
        return set(values)


class ArchivePolicy(Base, GnocchiBase, archive_policy.ArchivePolicy):
    __tablename__ = 'archive_policy'

    name = sqlalchemy.Column(sqlalchemy.String(255), primary_key=True)
    back_window = sqlalchemy.Column(sqlalchemy.Integer, nullable=False)
    definition = sqlalchemy.Column(ArchivePolicyDefinitionType, nullable=False)
    # TODO(jd) Use an array of string instead, PostgreSQL can do that
    aggregation_methods = sqlalchemy.Column(SetType,
                                            nullable=False)


class Metric(Base, GnocchiBase, indexer.Metric):
    __tablename__ = 'metric'
    __table_args__ = (
        sqlalchemy.Index('ix_metric_status', 'status'),
        sqlalchemy.UniqueConstraint("resource_id", "name",
                                    name="uniq_metric0resource_id0name"),
        COMMON_TABLES_ARGS,
    )

    id = sqlalchemy.Column(sqlalchemy_utils.UUIDType(),
                           primary_key=True)
    archive_policy_name = sqlalchemy.Column(
        sqlalchemy.String(255),
        sqlalchemy.ForeignKey(
            'archive_policy.name',
            ondelete="RESTRICT",
            name="fk_metric_ap_name_ap_name"),
        nullable=False)
    archive_policy = sqlalchemy.orm.relationship(ArchivePolicy, lazy="joined")
    creator = sqlalchemy.Column(sqlalchemy.String(255))
    resource_id = sqlalchemy.Column(
        sqlalchemy_utils.UUIDType(),
        sqlalchemy.ForeignKey('resource.id',
                              ondelete="SET NULL",
                              name="fk_metric_resource_id_resource_id"))
    name = sqlalchemy.Column(sqlalchemy.String(255))
    unit = sqlalchemy.Column(sqlalchemy.String(31))
    status = sqlalchemy.Column(sqlalchemy.Enum('active', 'delete',
                                               name="metric_status_enum"),
                               nullable=False,
                               server_default='active')

    def jsonify(self):
        d = {
            "id": self.id,
            "creator": self.creator,
            "name": self.name,
            "unit": self.unit,
        }
        unloaded = sqlalchemy.inspect(self).unloaded
        if 'resource' in unloaded:
            d['resource_id'] = self.resource_id
        else:
            d['resource'] = self.resource
        if 'archive_policy' in unloaded:
            d['archive_policy_name'] = self.archive_policy_name
        else:
            d['archive_policy'] = self.archive_policy

        if self.creator is None:
            d['created_by_user_id'] = d['created_by_project_id'] = None
        else:
            d['created_by_user_id'], _, d['created_by_project_id'] = (
                self.creator.partition(":")
            )

        return d

    def __eq__(self, other):
        # NOTE(jd) If `other` is a SQL Metric, we only compare
        # archive_policy_name, and we don't compare archive_policy that might
        # not be loaded. Otherwise we fallback to the original comparison for
        # indexer.Metric.
        return ((isinstance(other, Metric)
                 and self.id == other.id
                 and self.archive_policy_name == other.archive_policy_name
                 and self.creator == other.creator
                 and self.name == other.name
                 and self.unit == other.unit
                 and self.resource_id == other.resource_id)
                or (indexer.Metric.__eq__(self, other)))

    __hash__ = indexer.Metric.__hash__


RESOURCE_TYPE_SCHEMA_MANAGER = resource_type.ResourceTypeSchemaManager(
    "gnocchi.indexer.sqlalchemy.resource_type_attribute")


class ResourceTypeAttributes(sqlalchemy_utils.JSONType):
    def process_bind_param(self, attributes, dialect):
        return super(ResourceTypeAttributes, self).process_bind_param(
            attributes.jsonify(), dialect)

    def process_result_value(self, value, dialect):
        attributes = super(ResourceTypeAttributes, self).process_result_value(
            value, dialect)
        return RESOURCE_TYPE_SCHEMA_MANAGER.attributes_from_dict(attributes)


class ResourceType(Base, GnocchiBase, resource_type.ResourceType):
    __tablename__ = 'resource_type'
    __table_args__ = (
        sqlalchemy.UniqueConstraint("tablename",
                                    name="uniq_resource_type0tablename"),
        COMMON_TABLES_ARGS,
    )

    name = sqlalchemy.Column(sqlalchemy.String(255), primary_key=True,
                             nullable=False)
    tablename = sqlalchemy.Column(sqlalchemy.String(35), nullable=False)
    attributes = sqlalchemy.Column(ResourceTypeAttributes)
    state = sqlalchemy.Column(sqlalchemy.Enum("active", "creating",
                                              "creation_error", "deleting",
                                              "deletion_error", "updating",
                                              "updating_error",
                                              name="resource_type_state_enum"),
                              nullable=False,
                              server_default="creating")
    updated_at = sqlalchemy.Column(types.TimestampUTC, nullable=False,
                                   # NOTE(jd): We would like to use
                                   # sqlalchemy.func.now, but we can't
                                   # because the type of PreciseTimestamp in
                                   # MySQL is not a Timestamp, so it would
                                   # not store a timestamp but a date as an
                                   # integer.
                                   default=lambda: utils.utcnow())

    def to_baseclass(self):
        cols = {}
        for attr in self.attributes:
            cols[attr.name] = sqlalchemy.Column(attr.satype,
                                                nullable=not attr.required)
        return type(str("%s_base" % self.tablename), (object, ), cols)


class ResourceJsonifier(indexer.Resource):
    def __str__(self):
        return str(self.jsonify())

    def jsonify(self, attrs=None):
        d = dict(self)
        del d['revision']
        if 'metrics' not in sqlalchemy.inspect(self).unloaded:
            d['metrics'] = dict((m.name, str(m.id))
                                for m in self.metrics)

        if self.creator is None:
            d['created_by_user_id'] = d['created_by_project_id'] = None
        else:
            d['created_by_user_id'], _, d['created_by_project_id'] = (
                self.creator.partition(":")
            )

        if attrs:
            return {key: val for key, val in d.items() if key in attrs}
        else:
            return d


class ResourceMixin(ResourceJsonifier):
    @declarative.declared_attr
    def __table_args__(cls):
        return (sqlalchemy.CheckConstraint(
            'started_at <= ended_at',
            name="ck_{}_started_before_ended".format(
                cls.__tablename__
            )
        ),
            COMMON_TABLES_ARGS
        )

    @declarative.declared_attr
    def type(cls):
        return sqlalchemy.Column(
            sqlalchemy.String(255),
            sqlalchemy.ForeignKey('resource_type.name',
                                  ondelete="RESTRICT",
                                  name="fk_%s_resource_type_name" %
                                  cls.__tablename__),
            nullable=False)

    creator = sqlalchemy.Column(sqlalchemy.String(255))
    started_at = sqlalchemy.Column(types.TimestampUTC, nullable=False,
                                   default=lambda: utils.utcnow())
    revision_start = sqlalchemy.Column(types.TimestampUTC, nullable=False,
                                       default=lambda: utils.utcnow())
    ended_at = sqlalchemy.Column(types.TimestampUTC)
    user_id = sqlalchemy.Column(sqlalchemy.String(255))
    project_id = sqlalchemy.Column(sqlalchemy.String(255))
    original_resource_id = sqlalchemy.Column(sqlalchemy.String(255),
                                             nullable=False)


class Resource(ResourceMixin, Base, GnocchiBase):
    __tablename__ = 'resource'
    _extra_keys = ['revision', 'revision_end']
    revision = -1
    id = sqlalchemy.Column(sqlalchemy_utils.UUIDType(),
                           primary_key=True)
    revision_end = None
    metrics = sqlalchemy.orm.relationship(
        Metric, backref="resource",
        primaryjoin="and_(Resource.id == Metric.resource_id, "
        "Metric.status == 'active')")

    def get_metric(self, metric_name):
        m = super(Resource, self).get_metric(metric_name)
        if m:
            if sqlalchemy.orm.session.object_session(self):
                # NOTE(jd) The resource is already loaded so that should not
                # trigger a SELECT
                m.resource
            return m


class ResourceHistory(ResourceMixin, Base, GnocchiBase):
    __tablename__ = 'resource_history'

    revision = sqlalchemy.Column(sqlalchemy.Integer, autoincrement=True,
                                 primary_key=True)
    id = sqlalchemy.Column(sqlalchemy_utils.UUIDType(),
                           sqlalchemy.ForeignKey(
                               'resource.id',
                               ondelete="CASCADE",
                               name="fk_rh_id_resource_id"),
                           nullable=False)
    revision_end = sqlalchemy.Column(types.TimestampUTC, nullable=False,
                                     default=lambda: utils.utcnow())
    metrics = sqlalchemy.orm.relationship(
        Metric, primaryjoin="Metric.resource_id == ResourceHistory.id",
        foreign_keys='Metric.resource_id')


class ResourceExt(object):
    """Default extension class for plugin

    Used for plugin that doesn't need additional columns
    """


class ResourceExtMixin(object):
    @declarative.declared_attr
    def __table_args__(cls):
        return (COMMON_TABLES_ARGS, )

    @declarative.declared_attr
    def id(cls):
        tablename_compact = cls.__tablename__
        if tablename_compact.endswith("_history"):
            tablename_compact = tablename_compact[:-6]
        return sqlalchemy.Column(
            sqlalchemy_utils.UUIDType(),
            sqlalchemy.ForeignKey(
                'resource.id',
                ondelete="CASCADE",
                name="fk_%s_id_resource_id" % tablename_compact,
                # NOTE(sileht): We use to ensure that postgresql
                # does not use AccessExclusiveLock on destination table
                use_alter=True),
            primary_key=True
        )


class ResourceHistoryExtMixin(object):
    @declarative.declared_attr
    def __table_args__(cls):
        return (COMMON_TABLES_ARGS, )

    @declarative.declared_attr
    def revision(cls):
        tablename_compact = cls.__tablename__
        if tablename_compact.endswith("_history"):
            tablename_compact = tablename_compact[:-6]
        return sqlalchemy.Column(
            sqlalchemy.Integer,
            sqlalchemy.ForeignKey(
                'resource_history.revision',
                ondelete="CASCADE",
                name="fk_%s_revision_rh_revision"
                % tablename_compact,
                # NOTE(sileht): We use to ensure that postgresql
                # does not use AccessExclusiveLock on destination table
                use_alter=True),
            primary_key=True
        )


class HistoryModelIterator(models.ModelIterator):
    def __next__(self):
        # NOTE(sileht): Our custom resource attribute columns don't
        # have the same name in database than in sqlalchemy model
        # so remove the additional "f_" for the model name
        n = next(self.i)
        model_attr = n[2:] if n[:2] == "f_" else n
        return model_attr, getattr(self.model, n)


class ArchivePolicyRule(Base, GnocchiBase):
    __tablename__ = 'archive_policy_rule'

    name = sqlalchemy.Column(sqlalchemy.String(255), primary_key=True)
    archive_policy_name = sqlalchemy.Column(
        sqlalchemy.String(255),
        sqlalchemy.ForeignKey(
            'archive_policy.name',
            ondelete="RESTRICT",
            name="fk_apr_ap_name_ap_name"),
        nullable=False)
    metric_pattern = sqlalchemy.Column(sqlalchemy.String(255), nullable=False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime
import numpy

from gnocchi.rest.aggregates.api import Grouper
from gnocchi.tests import base
from unittest import mock


RETRIEVE_RESOURCES_HISTORY = \
    "gnocchi.rest.aggregates.api.Grouper.retrieve_resources_history"
API_AGGREGATE = \
    "gnocchi.rest.aggregates.api.AggregatesController._get_measures_by_name"


class Resource(object):
    def __init__(self, id, flavor_name, host, display_name,
                 revision_start=datetime.datetime(2020, 3, 10, 10, 0, 0, 0)):
        self.id = id
        self.flavor_name = flavor_name
        self.host = host
        self.display_name = display_name
        self.revision_start = revision_start
        self.revision_end = None


class ResourceHistory(object):
    def __init__(self, resources):
        self.resources = resources
        self.history = list(resources)

    def update_resource(self, date, update, id):
        resource = list(filter(lambda r: r.id == id, self.resources))[0]
        new_resource = Resource(resource.id, resource.flavor_name,
                                resource.host, resource.display_name, date)

        for k, v in update.items():
            new_resource.__setattr__(k, v)
        self.history.append(new_resource)

    def get_history_as_dict(self):
        self.history.sort(key=lambda r: r.revision_start)
        ids = set(map(lambda h: h.id, self.history))
        for id in ids:
            resources_with_id = [h for h in self.history if h.id == id]
            for i, resource in enumerate(resources_with_id[:-1]):
                resource.revision_end = resources_with_id[i + 1].revision_start

        return list(map(lambda x: x.__dict__, self.history))


class TestScenario(object):

    @mock.patch(RETRIEVE_RESOURCES_HISTORY)
    @mock.patch(API_AGGREGATE)
    def __init__(self, mock_measure, mock_history, test_input,
                 expected_output):
        self.test_input = test_input
        self.expected_output = expected_output
        self.result = None
        self.grouper = Grouper(test_input['test_group_by'],
                               test_input['test_start'],
                               test_input['test_end'],
                               {'operations': 'operation'},
                               None, None, None, None, None,
                               None, None)
        history = self.create_test_scenario()
        mock_history.side_effect = history.get_history_as_dict
        mock_measure.side_effect = get_metric
        self.execute()

    def execute(self):
        self.result = self.grouper.get_grouped_measures()

    def validate_scenario(self):
        all_groups = list(map(lambda o: o['group'], self.expected_output))
        all_response_groups = list(map(lambda r: r['group'], self.result))
        assert all_groups == all_response_groups
        for r in self.result:
            for out in self.expected_output:
                if out['group'] == r['group']:
                    for date, val in out['measures'].items():
                        aggregated = r['measures']['measures']['aggregated']
                        assert len(aggregated) == len(out['measures'])
                        for dat, gran, value in aggregated:
                            if str(dat) == date:
                                assert val == value

    def create_test_scenario(self):
        scenario = self.test_input
        resource_history = ResourceHistory(scenario['resource'])
        for update in scenario['updates']:
            resource_history.update_resource(update['event_time'],
                                             update['event_update'],
                                             update['id'])

        return resource_history


def get_metric(*args, **kwargs):
    start = args[3]
    end = args[4]
    if not end:
        end = numpy.datetime64('2020-03-10T12:00:00Z')
    ts = (start - numpy.datetime64(
        '1970-01-01T00:00:00Z')) / numpy.timedelta64(1, 's')
    current = numpy.datetime64(
        datetime.datetime.utcfromtimestamp(ts - (ts % 3600)))
    to_return = []
    while current < end:
        to_return.append((
            current,
            numpy.timedelta64(3600000000000, 'ns'),
            100
        ))
        current += numpy.timedelta64(3600, 's')

    return {'measures': {'aggregated': to_return}}


class TestGroupMeasuresWithHistory(base.BaseTestCase):

    def test_group_measures_changing_over_two_hours_with_more_than_one_resource(self):
        test_input = {
            'test_start': datetime.datetime(2020, 3, 10, 10, 0, 0, 0),
            'test_end': datetime.datetime(2020, 3, 10, 12, 0, 0, 0),
            'test_group_by': ['flavor_name', 'id'],
            'resource': [Resource(id=1, flavor_name="1gb-mem", host="192.168.0.1",
                                  display_name="My_VM",
                                  revision_start=datetime.datetime(
                                      2020, 3, 10, 9, 0, 0, 0)),
                         Resource(id=2, flavor_name="1gb-mem", host="192.168.0.1",
                                  display_name="My_VM",
                                  revision_start=datetime.datetime(
                                      2020, 3, 10, 8, 0, 0, 0))
                         ],
            'updates': [
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 10, 30, 0, 0),
                    'event_update': {'flavor_name': "2gb-mem"}
                },
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 11, 30, 0, 0),
                    'event_update': {'flavor_name': "1gb-mem"}
                },
                {
                    'id': 2,
                    'event_time': datetime.datetime(2020, 3, 10, 11, 0, 0, 0),
                    'event_update': {'flavor_name': "2gb-mem"}
                },
            ]
        }
        expected_output = [
            {
                'group': {'flavor_name': '1gb-mem', 'id': 1},
                'measures': {
                    '2020-03-10T10:00:00.000000': 50,
                    '2020-03-10T11:00:00.000000': 50
                }
            },
            {
                'group': {'flavor_name': '1gb-mem', 'id': 2},
                'measures': {
                    '2020-03-10T10:00:00.000000': 100
                }
            },
            {
                'group': {'flavor_name': '2gb-mem', 'id': 1},
                'measures': {
                    '2020-03-10T10:00:00.000000': 50,
                    '2020-03-10T11:00:00.000000': 50
                }
            },
            {
                'group': {'flavor_name': '2gb-mem', 'id': 2},
                'measures': {
                    '2020-03-10T11:00:00.000000': 100
                }
            },
        ]
        scenario = TestScenario(test_input=test_input,
                                expected_output=expected_output)
        scenario.validate_scenario()

    def test_group_measures_changing_over_two_hours(self):
        test_input = {
            'test_start': datetime.datetime(2020, 3, 10, 10, 0, 0, 0),
            'test_end': datetime.datetime(2020, 3, 10, 12, 0, 0, 0),
            'test_group_by': ['flavor_name'],
            'resource': [Resource(id=1, flavor_name="1gb-mem", host="192.168.0.1",
                                  display_name="My_VM")],
            'updates': [
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 10, 30, 0, 0),
                    'event_update': {'flavor_name': "2gb-mem"}
                },
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 11, 30, 0, 0),
                    'event_update': {'flavor_name': "1gb-mem"}
                }
            ]
        }
        expected_output = [
            {
                'group': {'flavor_name': '1gb-mem'},
                'measures': {
                    '2020-03-10T10:00:00.000000': 50,
                    '2020-03-10T11:00:00.000000': 50
                }
            },
            {
                'group': {'flavor_name': '2gb-mem'},
                'measures': {
                    '2020-03-10T10:00:00.000000': 50,
                    '2020-03-10T11:00:00.000000': 50
                }
            }
        ]
        scenario = TestScenario(test_input=test_input,
                                expected_output=expected_output)
        scenario.validate_scenario()

    def test_group_measures_changing_flavors_in_a_hour(self):
        test_input = {
            'test_start': datetime.datetime(2020, 3, 10, 10, 0, 0, 0),
            'test_end': datetime.datetime(2020, 3, 10, 12, 0, 0, 0),
            'test_group_by': ['flavor_name'],
            'resource': [Resource(id=1, flavor_name="1gb-mem", host="192.168.0.1",
                                  display_name="My_VM")],
            'updates': [
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 10, 30, 0, 0),
                    'event_update': {'flavor_name': "2gb-mem"}
                },
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 10, 45, 0, 0),
                    'event_update': {'flavor_name': "1gb-mem"}
                }
            ]
        }
        expected_output = [
            {
                'group': {'flavor_name': '1gb-mem'},
                'measures': {
                    '2020-03-10T10:00:00.000000': 75,
                    '2020-03-10T11:00:00.000000': 100
                }
            },
            {
                'group': {'flavor_name': '2gb-mem'},
                'measures': {
                    '2020-03-10T10:00:00.000000': 25
                }
            }
        ]
        scenario = TestScenario(test_input=test_input,
                                expected_output=expected_output)
        scenario.validate_scenario()

    def test_group_measures_many_flavors_changes_in_a_hour(self):
        test_input = {
            'test_start': datetime.datetime(2020, 3, 10, 10, 0, 0, 0),
            'test_end': datetime.datetime(2020, 3, 10, 12, 0, 0, 0),
            'test_group_by': ['flavor_name'],
            'resource': [Resource(id=1, flavor_name="1gb-mem", host="192.168.0.1",
                                  display_name="My_VM")],
            'updates': [
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 10, 10, 0, 0),
                    'event_update': {'flavor_name': "2gb-mem"}
                },
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 10, 20, 0, 0),
                    'event_update': {'flavor_name': "3gb-mem"}
                },
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 10, 30, 0, 0),
                    'event_update': {'flavor_name': "4gb-mem"}
                },
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 10, 40, 0, 0),
                    'event_update': {'flavor_name': "1gb-mem"}
                },
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 10, 45, 0, 0),
                    'event_update': {'flavor_name': "2gb-mem"}
                },
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 10, 55, 0, 0),
                    'event_update': {'flavor_name': "3gb-mem"}
                }
            ]
        }
        expected_output = [
            {
                'group': {'flavor_name': '1gb-mem'},
                'measures': {
                    '2020-03-10T10:00:00.000000': 24.999999999999996
                }
            },
            {
                'group': {'flavor_name': '2gb-mem'},
                'measures': {
                    '2020-03-10T10:00:00.000000': 33.33333333333333
                }
            },
            {
                'group': {'flavor_name': '3gb-mem'},
                'measures': {
                    '2020-03-10T10:00:00.000000': 24.999999999999996,
                    '2020-03-10T11:00:00.000000': 100,
                }
            },
            {
                'group': {'flavor_name': '4gb-mem'},
                'measures': {
                    '2020-03-10T10:00:00.000000': 16.666666666666686
                }
            }
        ]
        scenario = TestScenario(test_input=test_input,
                                expected_output=expected_output)
        scenario.validate_scenario()

    def test_group_measures_multiple_metadata_changed(self):
        test_input = {
            'test_start': datetime.datetime(2020, 3, 10, 10, 0, 0, 0),
            'test_end': datetime.datetime(2020, 3, 10, 12, 0, 0, 0),
            'test_group_by': ['flavor_name', 'display_name'],
            'resource': [Resource(id=1, flavor_name="1gb-mem", host="192.168.0.1",
                                  display_name="My_VM")],
            'updates': [
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 10, 30, 0, 0),
                    'event_update': {'flavor_name': "2gb-mem"}
                },
                {
                    'id': 1,
                    'event_time': datetime.datetime(2020, 3, 10, 10, 45, 0, 0),
                    'event_update': {'display_name': "Not_My_VM",
                                     'flavor_name': "2gb-mem"}
                }
            ]
        }
        expected_output = [
            {
                'group': {'flavor_name': '1gb-mem',
                          'display_name': 'My_VM'},
                'measures': {
                    '2020-03-10T10:00:00.000000': 50
                }
            },
            {
                'group': {'flavor_name': '2gb-mem',
                          'display_name': 'My_VM'},
                'measures': {
                    '2020-03-10T10:00:00.000000': 25
                }
            },
            {
                'group': {'flavor_name': '2gb-mem',
                          'display_name': 'Not_My_VM'},
                'measures': {
                    '2020-03-10T10:00:00.000000': 25,
                    '2020-03-10T11:00:00.000000': 100,
                }
            }
        ]
        scenario = TestScenario(test_input=test_input,
                                expected_output=expected_output)
        scenario.validate_scenario()



# -*- encoding: utf-8 -*-
#
# Copyright © 2016-2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import daiquiri
import fnmatch
import itertools

import numpy
import pecan
from pecan import rest
import pyparsing
import voluptuous

from gnocchi import indexer
from gnocchi.rest.aggregates import exceptions
from gnocchi.rest.aggregates import operations as agg_operations
from gnocchi.rest.aggregates import processor
from gnocchi.rest import api
from gnocchi import storage
from gnocchi import utils

LOG = daiquiri.getLogger(__name__)


def _OperationsSubNodeSchema(v):
    return OperationsSubNodeSchema(v)


def MetricSchema(v):
    """metric keyword schema

    It could be:

    ["metric", "metric-ref", "aggregation"]

    or

    ["metric, ["metric-ref", "aggregation"], ["metric-ref", "aggregation"]]
    """
    if not isinstance(v, (list, tuple)):
        raise voluptuous.Invalid("Expected a tuple/list, got a %s" % type(v))
    elif not v:
        raise voluptuous.Invalid("Operation must not be empty")
    elif len(v) < 2:
        raise voluptuous.Invalid("Operation need at least one argument")
    elif v[0] != u"metric":
        # NOTE(sileht): this error message doesn't looks related to "metric",
        # but because that the last schema validated by voluptuous, we have
        # good chance (voluptuous.Any is not predictable) to print this
        # message even if it's an other operation that invalid.
        raise voluptuous.Invalid("'%s' operation invalid" % v[0])

    return [u"metric"] + voluptuous.Schema(voluptuous.Any(
        voluptuous.ExactSequence([str, str]),
        voluptuous.All(
            voluptuous.Length(min=1),
            [voluptuous.ExactSequence([str, str])],
        )), required=True)(v[1:])


OperationsSchemaBase = [
    MetricSchema,
    voluptuous.ExactSequence(
        [voluptuous.Any(*list(
            agg_operations.ternary_operators.keys())),
         _OperationsSubNodeSchema, _OperationsSubNodeSchema,
         _OperationsSubNodeSchema]
    ),
    voluptuous.ExactSequence(
        [voluptuous.Any(*list(
            agg_operations.binary_operators.keys())),
         _OperationsSubNodeSchema, _OperationsSubNodeSchema]
    ),
    voluptuous.ExactSequence(
        [voluptuous.Any(*list(
            agg_operations.ternary_operators.keys())),
         _OperationsSubNodeSchema, _OperationsSubNodeSchema]
    ),
    voluptuous.ExactSequence(
        [voluptuous.Any(*list(
            agg_operations.unary_operators.keys())),
         _OperationsSubNodeSchema]
    ),
    voluptuous.ExactSequence(
        [voluptuous.Any(*list(
            agg_operations.unary_operators_with_timestamps.keys())),
         _OperationsSubNodeSchema]
    ),
    voluptuous.ExactSequence(
        [u"aggregate",
         voluptuous.Any(*list(agg_operations.AGG_MAP.keys())),
         _OperationsSubNodeSchema]
    ),
    voluptuous.ExactSequence(
        [u"resample",
         voluptuous.Any(*list(agg_operations.AGG_MAP.keys())),
         utils.to_timespan, _OperationsSubNodeSchema]
    ),
    voluptuous.ExactSequence(
        [u"rolling",
         voluptuous.Any(*list(agg_operations.AGG_MAP.keys())),
         voluptuous.All(
             voluptuous.Coerce(int),
             voluptuous.Range(min=1),
         ),
         _OperationsSubNodeSchema]
    )
]


OperationsSubNodeSchema = voluptuous.Schema(voluptuous.Any(*tuple(
    OperationsSchemaBase + [voluptuous.Coerce(float)]
)), required=True)


def OperationsSchema(v):
    if isinstance(v, str):
        try:
            v = pyparsing.OneOrMore(
                pyparsing.nestedExpr()).parseString(v).asList()[0]
        except pyparsing.ParseException as e:
            api.abort(400, {"cause": "Invalid operations",
                            "reason": "Fail to parse the operations string",
                            "detail": str(e)})
    return voluptuous.Schema(voluptuous.Any(*OperationsSchemaBase),
                             required=True)(v)


class ReferencesList(list):
    "A very simplified OrderedSet with list interface"

    def append(self, ref):
        if ref not in self:
            super(ReferencesList, self).append(ref)

    def extend(self, refs):
        for ref in refs:
            self.append(ref)


def extract_references(nodes):
    references = ReferencesList()
    if nodes[0] == "metric":
        if isinstance(nodes[1], list):
            for subnodes in nodes[1:]:
                references.append(tuple(subnodes))
        else:
            references.append(tuple(nodes[1:]))
    else:
        for subnodes in nodes[1:]:
            if isinstance(subnodes, list):
                references.extend(extract_references(subnodes))
    return references


def get_measures_or_abort(references, operations, start,
                          stop, granularity, needed_overlap, fill):
    try:
        return processor.get_measures(
            pecan.request.storage,
            references,
            operations,
            start, stop,
            granularity, needed_overlap, fill)
    except exceptions.UnAggregableTimeseries as e:
        api.abort(400, e)
    # TODO(sileht): We currently got only one metric for these exceptions but
    # we can improve processor to returns all missing metrics at once, so we
    # returns a list for the future
    except storage.MetricDoesNotExist as e:
        api.abort(404, {"cause": "Unknown metrics",
                        "detail": [str(e.metric.id)]})
    except storage.AggregationDoesNotExist as e:
        api.abort(404, {"cause": "Metrics with unknown aggregation",
                        "detail": [(str(e.metric.id), e.method)]})


class MeasureGroup(object):
    def __init__(self, group_key, resources):
        self.resources = self.join_sequential_groups(resources)
        self.group_key = dict(group_key)
        self.measures = []
        self.references = None

    def add_measures(self, aggregated_measures, date_map, start, stop,
                     details):
        measures = aggregated_measures['measures']['aggregated']
        if details:
            self.references = aggregated_measures['references']
        for measure in measures:
            self.add_measure(list(measure), date_map, start, stop)

    def add_measure(self, measure, date_map, start, stop):
        measure = Measure(*(measure + [start, stop]))
        if not date_map.get(measure.date, None):
            date_map[measure.date] = []
        date_map[measure.date].append(measure)
        self.measures.append(measure)

    def join_sequential_groups(self, group):
        group.sort(key=lambda key: key['revision_start'])
        new_group = []
        last_it = None
        for it in group:
            if last_it and it['revision_start'] == last_it['revision_end']:
                last_it['revision_end'] = it['revision_end']
                continue

            last_it = it
            new_group.append(it)

        return new_group

    def sum_groups_same_time_values(self):
        to_remove = []
        last_visited = None
        for measure in self.measures:
            if last_visited and last_visited.date == measure.date:
                to_remove.append(last_visited)
                measure.value += last_visited.value

            last_visited = measure

        self.measures = [m for m in self.measures if m not in to_remove]


class Measure(object):
    def __init__(self, date, granularity, value, start, stop):
        self.date = date
        self.granularity = granularity
        self.value = value
        self.start = (numpy.datetime64(start) if start
                      else date)
        measure_next_window_measurement = self.date + self.granularity
        self.stop = (numpy.datetime64(stop) if stop
                     else measure_next_window_measurement)
        measure_expected_beg = max(self.start, self.date)
        measure_expected_end = min(self.stop, measure_next_window_measurement)
        measure_delta = measure_expected_end - measure_expected_beg
        measure_delta_ns = measure_delta.astype('timedelta64[ns]')
        self.usage_coefficient = (measure_delta_ns.astype(float) /
                                  self.granularity.astype(float))


class Grouper(object):
    def __init__(self, groups, start, end, body, sorts, attr_filter,
                 references, granularity, needed_overlap, fill, details):
        self.groups = groups
        self.start = start
        self.end = end
        self.body = body
        self.sorts = sorts
        self.attr_filter = attr_filter
        self.references = references
        self.granularity = granularity
        self.needed_overlap = needed_overlap
        self.fill = fill
        self.details = details
        self.measures_date_map = {}
        self.grouped_response = None

    def create_history_period_filter(self):
        period_filter = {
            "and": [
                {
                    "<": {'revision_start': self.end}
                },
                {
                    "or": [
                        {">=": {'revision_end': self.start}},
                        {"=": {'revision_end': None}}
                    ]
                }
            ]
        }
        if not (self.start and self.end):
            return self.attr_filter

        if not self.attr_filter:
            return period_filter

        return {"and": [period_filter, self.attr_filter]}

    def retrieve_resources_history(self):
        return pecan.request.indexer.list_resources(
            self.body["resource_type"],
            attribute_filter=self.create_history_period_filter(),
            history=True,
            sorts=self.sorts)

    def get_grouped_measures(self):
        resources = self.retrieve_resources_history()
        self.grouped_response = self.get_measures(self.group(resources))
        response = self.format_response()
        LOG.debug("[ Resources History: %s ]", resources)
        LOG.debug("[ Response: %s ]", response)
        return response

    def group(self, to_group):
        to_group.sort(key=lambda g: g['revision_start'])
        is_first = True
        for value in to_group:
            self.truncate_resource_time_window(value, is_first=is_first)
            is_first = False
        to_group.sort(key=lambda x: tuple((attr, str(x[attr] or ''))
                                          for attr in self.groups))
        grouped_values = \
            itertools.groupby(
                to_group,
                lambda x: tuple((attr, x[attr]) for attr in self.groups))

        grouped = []
        for key, values in grouped_values:
            resources = []
            for value in values:
                resources.append(value)
            grouped.append(MeasureGroup(key, resources))

        return grouped

    def truncate_resource_time_window(self, value, is_first=False):
        if is_first:
            value['revision_start'] = self.start
        elif self.start:
            if value['revision_start']:
                revision_start = numpy.datetime64(value['revision_start'])
                value['revision_start'] = max(revision_start, self.start)
            else:
                value['revision_start'] = self.start

        if self.end:
            if value['revision_end']:
                revision_end = numpy.datetime64(value['revision_end'])
                value['revision_end'] = min(revision_end, self.end)
            else:
                value['revision_end'] = self.end

    def get_measures(self, groups):
        for group in groups:
            date_map = self.get_date_map(group)
            for resource in group.resources:
                start = numpy.datetime64(resource['revision_start']) \
                    if resource['revision_start'] else None
                stop = numpy.datetime64(resource['revision_end']) \
                    if resource['revision_end'] else None

                LOG.debug("[ Collecting measures from %s to %s ]",
                          start, stop)

                try:
                    measure = AggregatesController._get_measures_by_name(
                        [resource], self.references, self.body["operations"],
                        start, stop, self.granularity, self.needed_overlap,
                        self.fill, details=self.details)
                    group.add_measures(measure, date_map, start, stop,
                                       self.details)
                except indexer.NoSuchMetric:
                    continue

        self.truncate_measures()
        self.sum_sequential_group_metrics(groups)
        return groups

    def get_date_map(self, group):
        if 'id' in group.group_key:
            return self.measures_date_map.setdefault(group.group_key['id'], {})

        return self.measures_date_map

    def sum_sequential_group_metrics(self, groups):
        for group in groups:
            group.sum_groups_same_time_values()

    def truncate_measures(self):
        for measures in self.measures_date_map.values():
            if isinstance(measures, list):
                self.truncate_measure(measures)
                continue

            for measure in measures.values():
                self.truncate_measure(measure)

    def truncate_measure(self, measures_list):
        if measures_list and len(measures_list) == 1:
            return
        measures_sum = 0
        for measure in measures_list[:-1]:
            measure_new_val = (measure.value *
                               measure.usage_coefficient)
            measures_sum += measure_new_val
            measure.value = measure_new_val
        last_measure = measures_list[-1]
        last_measure.value = abs(measures_sum - last_measure.value)

    def format_response(self):
        measures_list = []
        for group in self.grouped_response:
            aggregated = []
            measures = {
                'measures': {'measures': {'aggregated': aggregated}},
                'group': group.group_key
            }
            if group.references:
                measures['measures']['references'] = group.references

            for measure in group.measures:
                aggregated.append((measure.date, measure.granularity,
                                   measure.value))

            if aggregated:
                measures_list.append(measures)

        return measures_list


def ResourceTypeSchema(resource_type):
    try:
        pecan.request.indexer.get_resource_type(resource_type)
    except indexer.NoSuchResourceType as e:
        api.abort(400, e)
    return resource_type


class AggregatesController(rest.RestController):

    FetchSchema = voluptuous.Any({
        "operations": OperationsSchema
    }, {
        "operations": OperationsSchema,
        "resource_type": ResourceTypeSchema,
        "search": voluptuous.Any(api.ResourceSearchSchema,
                                 api.QueryStringSearchAttrFilter.parse),
    })

    @pecan.expose("json")
    def post(self, start=None, stop=None, granularity=None,
             needed_overlap=None, fill=None, groupby=None, **kwargs):

        use_history = api.get_bool_param('use_history', kwargs)
        details = api.get_bool_param('details', kwargs)

        if fill is None and needed_overlap is None:
            fill = "dropna"
        start, stop, granularity, needed_overlap, fill = api.validate_qs(
            start, stop, granularity, needed_overlap, fill)

        body = api.deserialize_and_validate(self.FetchSchema)

        references = extract_references(body["operations"])
        if not references:
            api.abort(400, {"cause": "Operations is invalid",
                            "reason": "At least one 'metric' is required",
                            "detail": body["operations"]})

        if "resource_type" in body:
            attr_filter = body["search"]
            policy_filter = (
                pecan.request.auth_helper.get_resource_policy_filter(
                    pecan.request, "search resource", body["resource_type"]))
            if policy_filter:
                if attr_filter:
                    attr_filter = {"and": [
                        policy_filter,
                        attr_filter
                    ]}
                else:
                    attr_filter = policy_filter

            groupby = sorted(set(api.arg_to_list(groupby)))
            sorts = groupby if groupby else api.RESOURCE_DEFAULT_PAGINATION
            results = []
            try:
                if not groupby:
                    resources = pecan.request.indexer.list_resources(
                        body["resource_type"],
                        attribute_filter=attr_filter,
                        sorts=sorts)
                    return self._get_measures_by_name(
                        resources, references, body["operations"], start, stop,
                        granularity, needed_overlap, fill, details=details)

                if use_history:
                    results = self.get_measures_grouping_with_history(
                        attr_filter, body, details, fill, granularity, groupby,
                        needed_overlap, references, sorts, start, stop)
                else:
                    resources = pecan.request.indexer.list_resources(
                        body["resource_type"],
                        attribute_filter=attr_filter,
                        sorts=sorts)
                    results = self.get_measures_grouping(
                        body, details, fill, granularity, needed_overlap,
                        references, resources, start, stop, groupby)

            except indexer.NoSuchMetric as e:
                api.abort(404, str(e))
            except indexer.IndexerException as e:
                api.abort(400, str(e))
            except Exception as e:
                LOG.exception(e)
                raise e

            if not results:
                all_metrics_not_found = list(set((m for (m, a) in references)))
                all_metrics_not_found.sort()
                api.abort(404, str(
                    indexer.NoSuchMetric(all_metrics_not_found)))
            return results

        else:
            try:
                metric_ids = set(str(utils.UUID(m))
                                 for (m, a) in references)
            except ValueError as e:
                api.abort(400, {"cause": "Invalid metric references",
                                "reason": str(e),
                                "detail": references})

            metrics = pecan.request.indexer.list_metrics(
                attribute_filter={"in": {"id": metric_ids}})
            missing_metric_ids = (set(metric_ids)
                                  - set(str(m.id) for m in metrics))
            if missing_metric_ids:
                api.abort(404, {"cause": "Unknown metrics",
                                "reason": "Provided metrics don't exists",
                                "detail": missing_metric_ids})

            number_of_metrics = len(metrics)
            if number_of_metrics == 0:
                return []

            for metric in metrics:
                api.enforce("get metric", metric)

            metrics_by_ids = dict((str(m.id), m) for m in metrics)
            references = [processor.MetricReference(metrics_by_ids[m], a)
                          for (m, a) in references]

            response = {
                "measures": get_measures_or_abort(
                    references, body["operations"],
                    start, stop, granularity, needed_overlap, fill)
            }
            if details:
                response["references"] = metrics

            return response

    def get_measures_grouping(self, body, details, fill, granularity,
                              needed_overlap, references, resources, start,
                              stop, groupby):
        def groupper(r):
            return tuple((attr, r[attr]) for attr in groupby)

        results = []
        for key, resources in itertools.groupby(resources, groupper):
            try:
                results.append({
                    "group": dict(key),
                    "measures": self._get_measures_by_name(
                        resources, references, body["operations"],
                        start, stop, granularity, needed_overlap, fill,
                        details=details)
                })
            except indexer.NoSuchMetric:
                pass
        return results

    def get_measures_grouping_with_history(self, attr_filter, body, details,
                                           fill, granularity, groupby,
                                           needed_overlap, references,
                                           sorts, start, stop):
        grouper = Grouper(groupby, start, stop, body,
                          sorts, attr_filter, references,
                          granularity, needed_overlap,
                          fill, details)
        return grouper.get_grouped_measures()

    @staticmethod
    def _get_measures_by_name(resources, metric_wildcards, operations,
                              start, stop, granularity, needed_overlap, fill,
                              details):

        references = []
        for r in resources:
            references.extend([
                processor.MetricReference(m, agg, r, wildcard)
                for wildcard, agg in metric_wildcards
                for m in r.metrics if fnmatch.fnmatch(m.name, wildcard)
            ])

        if not references:
            all_metrics_not_found = list(
                set((m for (m, a) in metric_wildcards)))
            all_metrics_not_found.sort()
            raise indexer.NoSuchMetric(all_metrics_not_found)

        response = {
            "measures": get_measures_or_abort(
                references, operations, start, stop, granularity,
                needed_overlap, fill)
        }
        if details:
            response["references"] = set((r.resource for r in references))
        return response



# -*- coding: utf-8 -*-
#
# Gnocchi documentation build configuration file
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import pkg_resources

# -- General configuration -----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'gnocchi.gendoc',
    'sphinxcontrib.httpdomain',
    'sphinx.ext.autodoc',
    'reno.sphinxext',
]

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Gnocchi'
copyright = 'The Gnocchi Developers'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
release = pkg_resources.get_distribution('gnocchi').version

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = []

# A list of warning types to suppress arbitrary warning messages.
suppress_warnings = ['ref.term']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'furo'

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/gnocchi-logo.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = '_static/gnocchi-icon.ico'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
